// ----------------------------------------------------------------------------
// Expand.java
// The "expand" command
// $Id: Expand.java,v 1.14 2009/12/03 15:43:07 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import fmrisc.ProofNavigator.*;
import fmrisc.Proving.*;
import fmrisc.Semantics.*;
import fmrisc.Syntax.*;

import java.io.PrintWriter;
import java.util.*;
import org.w3c.dom.*;

/**
 * 
 * Command "expand": expand a definition.
 *
 */
public final class Expand extends ProofCommandBase
{
  public final static String Name = "expand";
  private Identifier[] idents; // sequence of value identifers
  private String[] labels;     // sequence of formula labels or null
  
  /***************************************************************************
   * Create an "expand" command.
   * @param idents the value identifiers whose definition are to be expanded.
   * @param labels the labels of the formulas in which the expansion shall
   *        take place (if null, all formulas are expanded).
   **************************************************************************/
  public Expand(Identifier[] idents, String[] labels)
  {
    super(Name);
    this.idents = idents;
    this.labels = labels;
  }
  
  /****************************************************************************
   * Process command "expand": expand value constant definitions in 
   * denoted formulas (respectively in all formulas of the state). 
   * @param state the proof state to which the command is applied
   * @return null or array of single new proof state.
   ***************************************************************************/
  protected ProofState[] processCore(ProofState state)
  {
    Environment env = state.getEnvironment();
    
    int n = idents.length;
    ValueSymbol[] symbols = new ValueSymbol[n];
    for (int i=0; i<n; i++)
    {
      symbols[i] = env.getValueSymbol(idents[i]);
      if (symbols[i] == null)
      {
        error("there is no value called " + idents[i].toString() +
        " visible in the current state");
        return null;
      }
      if (symbols[i].getValue() == null)
      {
        error("value " + idents[i].toString() + ":" + 
            symbols[i].getType().toString() +
        " has no definition");
        return null;
      }
      // replace identifier by symbol identifier for
      // later collection of symbols referenced in proof
      idents[i] = Construct.reference(symbols[i]);
    }
    
    if (labels != null)
    {
      int m = labels.length;
      for (int i=0; i<m; i++)
      {
        if (state.getFormula(labels[i]) == null)
        {
          error("there is no formula with label [" + labels[i] + 
            "] in the current state");
          return null;
        }
      }
    }
    
    Formula[] assumptions = state.getAssumptions();
    Formula[] goals = state.getGoals();
    ProofState state0 = new ProofState(state, env);
    Formula[] assumptions0 = expand(state0, symbols, assumptions, labels);
    if (assumptions0 == null) return null;
    state0.setAssumptions(assumptions0);
    Formula[] goals0 = expand(state0, symbols, goals, labels);
    if (goals0 == null) return null;
    state0.setGoals(goals0);
    return new ProofState[] { state0 };
  }
  
  /***************************************************************************
   * Expand definition of constants denoted by symbols.
   * @param state the state to which the resulting formulas shall belong.
   * @param symbols the constant symbols.
   * @param formulas the formulas.
   * @param labels the labels of those formulas where the expansion shall
   *        take place (null, if all formulas are expanded).
   * @return copy of formulas in state with expansions performed
   *         (null, if an error occured)
   **************************************************************************/
  private Formula[] expand(ProofState state, ValueSymbol[] symbols,
    Formula[] formulas, String[] labels)
  {
    List slist = null;
    if (labels != null) slist = Arrays.asList(labels);
    int n = formulas.length;
    Formula[] formulas0 = new Formula[n];
    for (int i=0; i<n; i++)
    {
      Formula f = formulas[i];
      if (slist != null && !slist.contains(f.getLabel()))
      {
        formulas0[i] = new Formula(state, f);
        continue;
      }
      int l = symbols.length;
      ValueDeclIdentifier[] idents0 = new ValueDeclIdentifier[l];
      Expression[] value0 = new Expression[l];
      for (int j=0; j<l; j++)
      {
        ValueSymbol symbol = symbols[j];
        idents0[j] = symbol.getValueDeclIdentifier();
        value0[j] = symbol.getValue();
      }
      Expression exp = f.getExpression();
      Expression exp0 = Substitute.substitute(exp, idents0, value0, true);
      formulas0[i] = new Formula(state, exp0, f.isGoal());
    }
    return formulas0;
  }
  
  /***************************************************************************
   * Set external representation of command in a DOM node.
   * @param node the node in which the representation is to be set
   *        (by default none, to be overridden by command class)
   **************************************************************************/
  protected void setRepresentation(Node node)
  {
    Proof proof = State.getProof();
    Node node0 = newNode("list");
    int n = idents.length;
    for (int i=0; i<n; i++)
    {
      Identifier ident = idents[i];
      if (proof != null) proof.referenced(ident);
      node0.appendChild(toNode(ident));
    }
    node.appendChild(node0);
    if (labels == null) return;
    Node node1 = newNode("list");
    n = labels.length;
    for (int i=0; i<n; i++)
    {
      Node node2 = newNode("label");
      node2.appendChild(newTextNode(labels[i]));
      node1.appendChild(node2);
    }
    node.appendChild(node1);
  }
  
  /***************************************************************************
   * Convert DOM node to command.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command newCommand(Node node)
  {
    NodeList children = node.getChildNodes();
    int l = children.getLength();
    if (l != 1 && l != 2)
    {
      State.setError("one or two children expected: " + node.toString());
      return null;
    }
    NodeList children0 = children.item(0).getChildNodes();
    int n = children0.getLength();
    Identifier[] idents = new Identifier[n];
    for (int i=0; i<n; i++)
    {
      idents[i] = toIdentifier(children0.item(i));
      if (idents[i] == null) return null;
    }
    if (l == 1) return new Expand(idents, null);
    NodeList children1 = children.item(1).getChildNodes();
    int m = children1.getLength();
    String[] labels = new String[m];
    for (int i=0; i<m; i++)
    {
      NodeList children2 =children1.item(i).getChildNodes();
      if (children2.getLength() != 1)
      {
        State.setError("one child expected: " + children1.item(i).toString());
        return null;
      }
      labels[i] = toString(children2.item(0));
      if (labels[i] == null) return null;
    }
    return new Expand(idents, labels);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
    if (idents != null)
    {
      out.print(" ");
      int n = idents.length;
      for (int i=0; i<n; i++)
      {
        idents[i].print(out);
        if (i<n-1) out.print(", ");
      }
    }
    if (labels != null)
    {
      out.print(" in ");
      int m = labels.length;
      for (int i=0; i<m; i++)
      {
        out.print(labels[i]);
        if (i<m-1) out.print(", ");
      }
    }
  }
}
