// ----------------------------------------------------------------------------
// Goal.java
// The "goal" command.
// $Id: Goal.java,v 1.3 2006/03/29 15:49:48 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import java.io.PrintWriter;
import java.util.*;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import fmrisc.ProofNavigator.State;
import fmrisc.Proving.*;

/*
 * command "goal": let a formula become the single goal.
 */
public final class Goal extends ProofCommandBase
{
  public final static String Name = "goal";
  private String label; // the label of the formula to become the goal
  
  /***************************************************************************
   * Create a "goal" command.
   * @param label the label of the formula to become the goal.
   **************************************************************************/
  public Goal(String label)
  {
    super(Name);
    this.label = label;
  }
  
  /****************************************************************************
   * process command "goal label": move all other formulas to assumption
   * positoin and goal as the only formula to goal position.
   * @param state the proof state to which the command is applied
   * @return null or array of single new proof state.
   ***************************************************************************/
  protected ProofState[] processCore(ProofState state)
  {
    ProofState state0 = new ProofState(state, state.getEnvironment());
    Vector assumptions0 = new Vector();
    Formula goal0 = null;
    
    Formula[] assumptions = state.getAssumptions();
    int n = assumptions.length;
    for (int i=0; i<n; i++)
    {
      Formula f = assumptions[i];
      if (goal0 == null && f.getLabel().equals(label))
        goal0 = new Formula(state0, 
          ProofUtil.negate(f.getExpression()), true);
      else
        assumptions0.add(new Formula(state0, f.getExpression(), false));
    }
    
    Formula[] goals = state.getGoals();
    n = goals.length;
    for (int i=0; i<n; i++)
    {
      Formula f = goals[i];
      if (goal0 == null && f.getLabel().equals(label))
        goal0 = new Formula(state0, f.getExpression(), true);
      else
        assumptions0.add(new Formula(state0, 
          ProofUtil.negate(f.getExpression()), false));
    }
    
    if (goal0 == null)
    {
      error("no formula with label " + label + " in current state");
      return null;
    }
    
    state0.setAssumptions(Formula.toArray(assumptions0));
    state0.setGoals(new Formula[] { goal0 });
    return new ProofState[] { state0 };
  }

  /***************************************************************************
   * Set external representation of command in a DOM node.
   * @param node the node in which the representation is to be set
   *        (by default none, to be overridden by command class)
   **************************************************************************/
  protected void setRepresentation(Node node)
  {
    node.appendChild(newTextNode(label));
  }
  
  /***************************************************************************
   * Convert DOM node to command.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command newCommand(Node node)
  {
    NodeList children = node.getChildNodes();
    if (children.getLength() != 1) 
    {
      State.setError("child expected: " + node.toString());
      return null;
    }
    String label = toString(children.item(0));
    if (label == null) return null;
    return new Goal(label);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
    out.print(" ");
    out.print(label);
  }
}
