// ----------------------------------------------------------------------------
// Option.java
// The "option" command.
// $Id: Option.java,v 1.1 2007/10/23 20:09:59 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import java.io.*;
import org.w3c.dom.*;

import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.SWT.*;
import fmrisc.Proving.*;

/**
 * 
 * The "option" command: set a system option.
 *
 */
public final class Option extends ProofCommandBase
{
  public final static String Name = "option";
  
  private String name;  // name of option
  private String value; // its value (may be null)
  
  /***************************************************************************
   * Create an "option" command.
   * @param name name of the option
   * @param value its value (may be null)
   **************************************************************************/
  public Option(String name, String value)
  {
    super(Name);
    this.name = name;
    this.value = value;
  }
  
  /****************************************************************************
   * override method to allow command also to be processed when not in proof
   ***************************************************************************/
  public void process()
  {
    if (State.getProof() == null)
    {
      processCore();
      MainSWT.updateMenu();
    }
    else
      super.process();
  }
  
  /****************************************************************************
   * Process command "option": set an option.
   * @return true iff everything went fine.
   ***************************************************************************/
  private boolean processCore()
  {
    if (name.equals("autosimp"))
    {
      String value0 = value;
      if (value0 == null) value0 = "true";
      if (value0.equals("true"))
      {
        if (State.useAutoSimplify())
        {
          error("automatic simplification already switched on");
          return false;
        }
        State.setAutoSimplify(true);
      }
      else if (value0.equals("false"))
      {
        if (!State.useAutoSimplify())
        {
          error("automatic simplification already switched off");
          return false;
        } 
        State.setAutoSimplify(false);
      }
      else
      {
        error("unsupported option value \"" + value0 + "\"");
        return false;
      }   
    }
    else
    {
      error("unknown option name \"" + name +"\"");
      return false;
    }
    return true;
  }
  
  /****************************************************************************
   * Process command "option": set an option.
   * @param state the proof state to which the command is applied
   * @return null or an array with a single state
   ***************************************************************************/
  protected ProofState[] processCore(ProofState state)
  {
    if (name.equals("autosimp"))
    {
      ProofState state0 = new ProofState(state, state.getEnvironment());
      state0.setAssumptions(state0.copy(state.getAssumptions()));
      state0.setGoals(state0.copy(state.getGoals()));
      String value0 = value;
      if (value0 == null) value0 = "true";
      if (value0.equals("true"))
      {
        if (state0.useAutoSimplify())
        {
          error("automatic simplification already switched on");
          return null;
        }
        state0.setAutoSimplify(true);
      }
      else if (value0.equals("false"))
      {
        if (!state0.useAutoSimplify())
        {
          error("automatic simplification already switched off");
          return null;
        } 
        state0.setAutoSimplify(false);
      }
      else
      {
        error("unsupported option value \"" + value0 + "\"");
        return null;
      }   
      return new ProofState[]{ state0 };
    }
    error("unknown option name \"" + name +"\"");
    return null;
  }
  
  /***************************************************************************
   * Set external representation of command in a DOM node.
   * @param node the node in which the representation is to be set
   *        (by default none, to be overridden by command class)
   **************************************************************************/
  protected void setRepresentation(Node node)
  {
    Node node0 = newNode("name");
    node0.appendChild(newTextNode(name));
    node.appendChild(node0);
    if (value == null) return;
    Node node1 = newNode("value");
    node1.appendChild(newTextNode(value));
    node.appendChild(node1);
  }
  
  /***************************************************************************
   * Convert DOM node to command.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command newCommand(Node node)
  {
    NodeList children = node.getChildNodes();
    int n = children.getLength();
    if (n == 0 || n > 2) 
    {
      State.setError("one or two children expected: " + node.toString());
      return null;
    }
    NodeList nnode = children.item(0).getChildNodes();
    if (nnode.getLength() != 1)
    {
      State.setError("one child expected: " + nnode.toString());
      return null;
    }
    String name = toString(nnode.item(0));
    if (n == 1) return new Option(name, null);
    NodeList vnode = children.item(1).getChildNodes();
    if (vnode.getLength() != 1)
    {
      State.setError("one child expected: " + vnode.toString());
      return null;
    }
    String value = toString(vnode.item(0));
    return new Option(name, value);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(Name);
    out.print(" ");
    out.print(name);
    if (value == null) return;
    out.print("=\"");
    out.print(value);
    out.print("\"");
  }
}