// ----------------------------------------------------------------------------
// Skolemize.java
// The "skolemize" command.
// $Id: Skolemize.java,v 1.7 2006/03/29 15:49:48 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import java.io.PrintWriter;

import org.w3c.dom.Node;

import fmrisc.Proving.*;
import fmrisc.Semantics.*;
import fmrisc.Syntax.*;

/*
 * command "skolemize": skolemize universal formulas in goal position and 
 * existential formulas in assumption position.
 */
public final class Skolemize extends ProofCommandBase 
{
  public final static String Name = "skolemize";
  
  /***************************************************************************
   * create a "skolemize" command
   **************************************************************************/
  public Skolemize()
  {
    super(Name);
  }
  
  /****************************************************************************
   * process command "skolemize": skolemize universal formulas in goal position 
   * and existential formulas in assumption position.
   * @param state the proof state to which the command is applied
   * @return null or array of single new proof state.
   ***************************************************************************/
  protected ProofState[] processCore(ProofState state)
  {
    Environment env = state.getEnvironment().cloneGlobal();
    ProofState state0 = new ProofState(state, env);
    UniqueNameTable skolemTable = state0.getSkolemTable();
    boolean changed = false;
    
    Formula[] assumptions = state.getAssumptions();
    int n = assumptions.length;
    Formula[] assumptions0 = new Formula[n];
    for (int i=0; i<n; i++)
    {
      Formula formula = assumptions[i];
      Expression exp = formula.getExpression();
      if (!(exp instanceof ExistsFormula))
      {
        assumptions0[i] = new Formula(state0, formula);
        continue;
      }
      changed = true;
      ExistsFormula f = (ExistsFormula)exp.instantiate();
      addConstants(env, f.getVariables(), skolemTable);
      assumptions0[i] = new Formula(state0, f.getBase(), false);
    } 
    
    Formula[] goals = state.getGoals();
    int m = goals.length;
    Formula[] goals0 = new Formula[m];
    for (int i=0; i<m; i++)
    {
      Formula formula = goals[i];
      Expression exp = formula.getExpression();
      if (!(exp instanceof ForallFormula))
      {
        goals0[i] = new Formula(state0, formula);
        continue;
      }
      changed = true;
      ForallFormula f = (ForallFormula)exp.instantiate();
      addConstants(env, f.getVariables(), skolemTable);
      goals0[i] = new Formula(state0, f.getBase(), true);
    }  
    
    if (!changed) return null;
    
    state0.setAssumptions(assumptions0);
    state0.setGoals(goals0);
    return new ProofState[] { state0 };
  }
  
  /***************************************************************************
   * add to environment uniquely named constants for all the denoted variables;
   * as a side effect, all references to these variables are replaced
   * by references to these constants.
   * @param env the environment to be extended
   * @param vars the variables for which constants are added
   * @param table the name table to be used for making the names unique
   **************************************************************************/
  private static void addConstants(Environment env, TypedIdentifier[] vars,
    UniqueNameTable table)
  {
    int n = vars.length;
    for (int i=0; i<n; i++)
    {
      ValueDeclIdentifier ident = vars[i].getIdentifier();
      ident.makeNameUnique(table);
      // make sure symbol is not treated as global in
      // collection of symbols referenced by proof
      ident.getSymbol().setDepth(1);
      env.put(ident);
    }
  }
  
  /***************************************************************************
   * Convert DOM node to command.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command newCommand(Node node)
  {
    return new Skolemize();
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
  }
}
