// ----------------------------------------------------------------------------
// Instantiating.java
// Extending a proof state by instantiating quantified formulas.
// $Id: Instantiating.java,v 1.10 2006/06/26 17:09:26 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Proving;

import java.util.*;
import fmrisc.ProofNavigator.*;
import fmrisc.Syntax.*;
import fmrisc.Semantics.*;

/**
 * 
 * Extending a proof state by instantiating quantified formulas.
 *
 */
public final class Instantiating
{
  // limit on number of variables instantiated in quantifier (-1: no limit)
  private static int limitVars = 0;
  
  // limit on number of instantiations in current result (-1: no limit)
  private static int limitInst = 0;
   
  // constants used for above values in state instantiation
  private final static int STATEVARS = -1;
  private final static int STATEINSTONE = 25;  // instantiations per formula
  private final static int STATEINSTALL = 200; // total number of instantiations
  
  // constants used for above values in instantiation of single formula
  private final static int FORMULAVARS = -1;
  private final static int FORMULAINST = 200;
  
  // constants used for above values in instantiation of multiple formulas
  private final static int MULTVARS = -1;
  private final static int MULTINSTONE = 50;  // instantiations per formula
  private final static int MULTINSTALL = 200; // total number of instantiations
  
  /****************************************************************************
   * Extend given proof state by instantiations of quantified formulas.
   * @param state an open proof state.
   * @return null or a proof state whose parent is the given state and
   * that is equivalent to the given state but with
   * instantiated versions of quantified formulas added such that
   * it is more likely that the state can be closed by a validity checker.
   ***************************************************************************/
  public static ProofState instantiate(ProofState state)
  {
    // set instantiation limits
    limitVars = STATEVARS;
    limitInst = STATEINSTONE;
    
    // first create a copy of the state
    Environment env = state.getEnvironment();
    ProofState state0 = new ProofState(state, env);

    // now determine the ground expressions in the state
    // and group them according to their types
    GroundExpressions exps = GroundExpressions.compute(state);
    Map texps = groupTypes(exps.getExpressions(), env);

    // collect instantiated assumptions here
    Vector result0 = new Vector();
    
    // instantiate type checking conditions
    // use low-level types for instantiation matching
    Checking.setWeakerTypes(true); 
    Iterator symbols = env.getValueSymbols().iterator();
    while (symbols.hasNext() && result0.size() < STATEINSTALL)
    {
      ValueSymbol symbol = (ValueSymbol)symbols.next();
      TypeExpression te = symbol.getClearType();
      if (te == null) continue;
      Expression exp = te.getExpression();
      if (exp == null || !(exp instanceof ForallFormula)) continue;
      Formula f = new Formula(state, exp, false);
      Vector r = new Vector();
      appendInstantiations(r, f, exps, texps);
      result0.addAll(r);
    }
    Checking.setWeakerTypes(false);
    
    // instantiate universal axioms
    Formula[] formulas = state.getAxioms();
    int n = formulas.length;
    for (int i=0; i<n && result0.size() < STATEINSTALL; i++)
    {
      Formula f = formulas[i];
      if (!(f.getExpression() instanceof ForallFormula)) continue;
      Vector r = new Vector();
      appendInstantiations(r, f, exps, texps); 
      result0.addAll(r);
    }
    
    // instantiate universal assumptions
    formulas = state.getAssumptions();
    n = formulas.length;
    for (int i=0; i<n && result0.size() < STATEINSTALL; i++)
    {
      Formula f = formulas[i];
      if (!(f.getExpression() instanceof ForallFormula)) continue;
      Vector r = new Vector();
      appendInstantiations(r, f, exps, texps); 
      result0.addAll(r);
    }
    
    // add instantiated assumptions
    state0.setAssumptions(
      state0.add(formulas, ProofUtil.toFormulaArray(result0)));
    
    // instantiate existential goals
    Vector result1 = new Vector();
    formulas = state.getGoals();
    n = formulas.length;
    for (int i=0; i<n && result0.size()+result1.size() < STATEINSTALL; i++)
    {
      Formula f = formulas[i];
      if (!(f.getExpression() instanceof ExistsFormula)) continue;
      Vector r = new Vector();
      appendInstantiations(r, f, exps, texps); 
      result1.addAll(r);
    }
    state0.setGoals(
      state0.add(formulas, ProofUtil.toFormulaArray(result1)));

    // return null or the new state
    if (result0.isEmpty() && result1.isEmpty()) 
      return null;
    else
      return state0;
  }
  
  /****************************************************************************
   * Extend proof state by instantiations of given formula
   * @param formula a formula
   * @return null or a proof state whose parent is the state of the formula and
   * that is equivalent to the given state but with
   * instantiated versions of the given formula added such that
   * it is more likely that the state can be closed by a validity checker.
   ***************************************************************************/
  public static ProofState instantiate(Formula formula)
  {
    // set instantiation limits
    limitVars = FORMULAVARS;
    limitInst = FORMULAINST;
    
    Expression exp = formula.getExpression();
    if ((formula.isGoal() && !(exp instanceof ExistsFormula)) ||
        (!formula.isGoal() && !(exp instanceof ForallFormula))) return null;
    
    // first create a copy of the state
    ProofState state = formula.getState();
    Environment env = state.getEnvironment();
    ProofState state0 = new ProofState(state, env);

    // now determine the ground expressions in the state
    // and group them according to their types
    GroundExpressions exps = GroundExpressions.compute(state);
    Map texps = groupTypes(exps.getExpressions(), env);
    
    // instantiate formula and add instantiations to state
    Vector result = new Vector();
    appendInstantiations(result, formula, exps, texps); 
    Formula[] formulas = ProofUtil.toFormulaArray(result);
    if (formula.isGoal())
    {
      state0.setAssumptions(state0.copy(state.getAssumptions()));
      state0.setGoals(state0.add(state.getGoals(), formulas));
    }
    else
    {
      state0.setAssumptions(state0.add(state.getAssumptions(), formulas));
      state0.setGoals(state0.copy(state.getGoals()));
    }
    if (result.isEmpty())
      return null;
    else
      return state0;
  }
  
  /****************************************************************************
   * Extend proof state by instantiations of given formulas
   * @param formulas a sequence of formulas in the same state
   * @return null or a proof state whose parent is the state of the formulas 
   * and that is equivalent to the given state but with
   * instantiated versions of the given formulas added such that
   * it is more likely that the state can be closed by a validity checker.
   ***************************************************************************/
  public static ProofState instantiate(Formula[] formulas)
  {
    // set instantiation limits
    limitVars = MULTVARS;
    limitInst = MULTINSTONE;
    
    // first create a copy of the state
    int n = formulas.length;
    if (n == 0) return null;
    ProofState state = formulas[0].getState();
    Environment env = state.getEnvironment();
    ProofState state0 = new ProofState(state, env);
    
    // now determine the ground expressions in the state
    // and group them according to their types
    GroundExpressions exps = GroundExpressions.compute(state);
    Map texps = groupTypes(exps.getExpressions(), env);
    
    // instantiate formula 
    Vector result0 = new Vector();
    Vector result1 = new Vector();
    for (int i=0; i<n && result0.size()+result1.size()<MULTINSTALL; i++)
    {
      Formula formula = formulas[i];
      Expression exp = formula.getExpression();
      if ((formula.isGoal() && !(exp instanceof ExistsFormula)) ||
         (!formula.isGoal() && !(exp instanceof ForallFormula))) continue;
    
      Vector r = new Vector();
      appendInstantiations(r, formula, exps, texps); 
      if (formula.isGoal())
        result1.addAll(r);
      else
        result0.addAll(r);
    }
    if (result0.isEmpty() && result1.isEmpty()) return null;
    
    // add instantiations to state
    Formula[] formulas0 = ProofUtil.toFormulaArray(result0);
    Formula[] formulas1 = ProofUtil.toFormulaArray(result1);
    state0.setAssumptions(state0.add(state.getAssumptions(), formulas0));
    state0.setGoals(state0.add(state.getGoals(), formulas1));
    return state0;
  }
    
  /***************************************************************************
   * Compute expressions according to their type.
   * @param c a collection of expressions.
   * @return a collection of (Type type, Vector exps) mappings
   * where each element of exps is an Expression object that has the 
   * denoted type, and the set of all expressions in all mappings
   * is the same as the expressions in c.
   **************************************************************************/
  private static Map groupTypes(Collection c, Environment env)
  {
    // we assume that type checking yields canonical types
    // which may be used as keys in the map table using the
    // basic equals() function in class Object.
    Map map = new HashMap();
    Environment oenv = State.getEnvironment();
    State.setEnvironment(env);
    Iterator iterator = c.iterator();
    while (iterator.hasNext())
    {
      Expression exp = (Expression)iterator.next();
      Type type = Checking.process(exp);
      Vector exps = (Vector)map.get(type);
      if (exps == null)
      {
        exps = new Vector();
        map.put(type, exps);
      }
      exps.addElement(exp);
    }
    State.setEnvironment(oenv);
    return map;
  }
  
  /****************************************************************************
   * Append to result vector interesting instantiations of quantified formula.
   * @param result the vector to which generated instantiations are appended.
   * @param formula a universal or existential formula to be instantiated.
   * @param exps all the ground expressions occuring in the proof state.
   * @param texps the ground expressions grouped according to their type.
   ***************************************************************************/
  private static void appendInstantiations(Vector result, 
    Formula formula, GroundExpressions exps, Map texps)
  {
    // determine all ground expressions occuring in formula
    QuantifiedExpression fexp = (QuantifiedExpression)formula.getExpression();
    // GroundExpressions fexps = GroundExpressions.compute(fexp);
    
    // compute all possible instantiations of formula
    Iterator iterator = allInstantiations(fexp, texps).iterator();
    
    // add every *interesting* instantiation to the result
    while (iterator.hasNext())
    {
      Expression iexp = (Expression)iterator.next();
      
      // compute all ground expressions in instantated formula
      Iterator iexps = 
        GroundExpressions.compute(iexp).getExpressions().iterator();

      // add instantiation to result, if some (new?) ground expression
      // also occurs in proof state
      while (iexps.hasNext() && result.size() < limitInst)
      {
        Expression e = (Expression)iexps.next();
        if (exps.has(e)) // && !fexps.has(e)
        {
          result.add(new Formula(formula.getState(), iexp, formula.isGoal()));
          break;
        }
      }
    }
  }
  
  /***************************************************************************
   * Compute all instantiations of quantified formula from a collection
   * of ground expressions.
   * @param formula a universally or existentially quantified formula
   * @param texps a map of (Type t, Vector v) pairs where all elements
   *        of v are ground expressions of type t.
   * @return the collection of all possible instantiations of formula
   *         with ground expressions from texps.
   **************************************************************************/
  private static Collection allInstantiations(
    QuantifiedExpression formula, Map texps)
  {
    Vector result = new Vector();
    TypedIdentifier[] vars = formula.getVariables();
    int n = vars.length;
    if (limitVars != -1 && n > limitVars) return result;
    ValueDeclIdentifier[] idents = new ValueDeclIdentifier[n];
    for (int i=0; i<n; i++) 
      idents[i] = vars[i].getIdentifier();
    Expression[] subst = new Expression[n];
    Expression[] cond = new Expression[n];
    appendInstantiations(result, idents, formula, texps, subst, cond, 0);
    return result;
  }
  
  /***************************************************************************
   * Append to result vector all instantiations of a formula
   * that are consistent with a given partial instantiation.
   * @param idents an array of identifiers.
   * @param formula a quantified formula whose *base* is to be instantiated.
   * @param texps a map of (Type t, Vector v) pairs where all elements
   *        of v are ground expressions of type t.
   * @param subst an expression array whose length equals the length 
   *        of idents and whose first n elements represent
   *        a partial instantiation for the corresponding elements of idents.
   * @param cond an expression array whose length equals the length 
   *        of idents and where each of the first n elements is either null
   *        or represents a side condition for the instantiation.
   * @param n the number of expressions in subst.
   ***************************************************************************/
  private static void appendInstantiations(Vector result,
    ValueDeclIdentifier[] idents, QuantifiedExpression formula, 
    Map texps, Expression[] subst, Expression[] cond, int n)
  {
    // base case
    if (idents.length == n)
    {
      // generate potential side conditions
      Expression condition = null;
      for (int i=0; i<n; i++)
      {
        if (cond[i]== null) continue;
        if (condition == null)
          condition = cond[i];
        else
          condition = Construct.andFormula(condition, cond[i]);
      }
      Expression f = formula.getBase();
      if (condition != null)
      {
        if (formula instanceof ForallFormula)
          f = Construct.impliesFormula(condition, f);
        else
          f = Construct.andFormula(condition, f);
      }
      result.addElement(Substitute.substitute(f, idents, subst, false));
      return;
    }
    
    // determine the type of the identifier to be instantiated next
    Type type = idents[n].getSymbol().getType();
    
    // instantiate boolean variable with TRUE and FALSE,
    // not with all ground formulas
    if (Checking.isBoolean(type))
    {
      subst[n] = Construct.logical(true);
      appendInstantiations(result, idents, formula, texps, subst, cond, n+1);
      subst[n] = Construct.logical(false);
      appendInstantiations(result, idents, formula, texps, subst, cond, n+1);
      return;
    }
      
    // we also may use type cleared from subtype for matching; in this case, 
    // we will generate corresponding side condition for instantiation below
    Type type1 = type;
    TypeExpression te = Checking.clearSubtypes(type, Construct.logical(true));
    if (te != null && te.getType() != null) type1 = te.getType();
    
    // instantiate with all ground expressions of appropriate type
    Iterator iterator = texps.entrySet().iterator();
    while (iterator.hasNext())
    {
      Map.Entry texp = (Map.Entry)iterator.next();
      Type type0 = (Type)texp.getKey();
      // first exact matching
      boolean matched = Checking.hasType(null, type0, type, false);
      if (!matched)
      {
        // if necessary, matching to cleared type
        if (type1 == type) continue;
        if (!Checking.hasType(null, type0, type1, false)) continue;
      }
      Vector sexps = (Vector)texp.getValue();
      int l = sexps.size();
      for (int i=0; i<l; i++)
      {
        subst[n] = (Expression)sexps.elementAt(i);
        cond[n] = null;
        if (!matched && type1 != type)
        {
          // have to generate side condition in instantiation
          TypeExpression te0 = Checking.clearSubtypes(type, subst[n]);
          if (te0 != null && te0.getExpression() != null)
            cond[n] = te0.getExpression();
        }
        appendInstantiations(result, idents, formula, texps, subst, cond, n+1);
      }
    }
  }
}
