// ----------------------------------------------------------------------------
// FreeVariables.java
// Compute the free variables of an expression.
// $Id: FreeVariables.java,v 1.2 2006/03/23 17:44:11 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Semantics;

import fmrisc.Syntax.*;
import java.util.*;

/**
 * 
 * Compute the free variables of an expression.
 *
 */
public final class FreeVariables extends ASTVisitorBase
{
  /***************************************************************************
   * Compute free variables of expression.
   * @param exp an expression.
   * @return a vector of ValueSymbol objects representing the free variables
   *         of the expression (each symbol occurs once in the vector).
   **************************************************************************/
  public static Vector compute(Expression exp)
  {
    FreeVariables fv = new FreeVariables();
    exp.accept(fv);
    return fv.getVariables();
  }
  
  /***************************************************************************
   * Compute free variables of type.
   * @param type a type.
   * @return a vector of ValueSymbol objects representing the free variables
   *         of the type (each symbol occurs once in the vector).
   **************************************************************************/
  public static Vector compute(Type type)
  {
    FreeVariables fv = new FreeVariables();
    type.accept(fv);
    return fv.getVariables();
  }
  
  /***************************************************************************
   * Compute free variables of expression.
   * @param exp an expression.
   * @param result a vector to which the ValueSymbol objects representing the 
   *        free variables of the expression are appended (each symbol will
   *        occur only once in result).
   **************************************************************************/
  public static void append(Expression exp, Vector result)
  {
    FreeVariables fv = new FreeVariables(result);
    exp.accept(fv);
  }
  
  // the vector holding the free variable symbols computed up to now
  private Vector free = null;
  
  /***************************************************************************
   * Create visitor with empty result vector.
   **************************************************************************/
  private FreeVariables()
  {
    this(new Vector());
  }
  
  /****************************************************************************
   * Create visitor with denoted result vector.
   * @param result the vector to which the free variables are added.
   ***************************************************************************/
  private FreeVariables(Vector result)
  {
    free = result;
  }
  
  /****************************************************************************
   * Get free variables computed by the expression visitor.
   * @return a vector of ValueSymbol objects holding the free variables.
   ***************************************************************************/
  private Vector getVariables()
  {
    return free;
  }
  
  /****************************************************************************
   * Add value symbol to list of free variables.
   * @param ident the reference whose symbol is added to list of free
   *        variables (if ident refers to value).
   * @return ident (signalling that no subtree remains to be visited).
   ***************************************************************************/
  protected AST visit(Reference ident) 
  { 
    Symbol symbol = ident.getSymbol();
    if (!(symbol instanceof ValueSymbol)) return ident;
    if (free.contains(symbol)) return ident;
    free.add(symbol);
    return ident;
  }
  
  /****************************************************************************
   * Process quantified expression, removing the variables bound by the 
   * quantifier from the list of free variables.
   * @param exp the quantified expression to be processed.
   * @return exp (signalling that no subtree remains to be visited).
   ***************************************************************************/
  protected AST visit(QuantifiedExpression exp)
  {
    TypedIdentifier[] variables = exp.getVariables();
    
    // process types since they may contain expressions
    int n = variables.length;
    for (int i=0; i<n; i++)
      variables[i].getType().accept(this);
    
    // process base expression
    exp.getBase().accept(this);
    
    // remove bound variables
    for (int i=0; i<n; i++)
      free.remove(variables[i].getIdentifier().getSymbol());
    
    // we are done
    return exp;
  }
  
  /****************************************************************************
   * Process let expression, removing the variables bound by the 
   * quantifier from the list of free variables.
   * @param exp the quantified expression to be processed.
   * @return exp (signalling that no subtree remains to be visited).
   ***************************************************************************/
  protected AST visit(LetExpression exp)
  {
    ValueDeclaration[] decls = exp.getDeclarations();
    
    // process types since they may contain expressions
    int n = decls.length;
    for (int i=0; i<n; i++)
    {
      Type type = decls[i].getType();
      if (type != null) type.accept(this);
      decls[i].getValue().accept(this);
    }
    
    // process base expression
    exp.getBase().accept(this);
    
    // remove bound variables
    for (int i=0; i<n; i++)
      free.remove(decls[i].getSymbol());
    
    // we are done
    return exp;
  }
}
