// ----------------------------------------------------------------------------
// Construct.java
// Construct abstract syntax trees.
// $Id: Construct.java,v 1.20 2006/03/01 17:10:00 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Syntax;

import fmrisc.Semantics.*;

/**
 * Construct abstract syntax trees.
 */
public final class Construct
{  
  /****************************************************************************
   * construct referenc with denoted name.
   * @param name the name of the reference
   * @return the reference
   ***************************************************************************/
  public static Reference reference(String name)
  {
    return new Reference(name);
  }
  
  /****************************************************************************
   * Construct referenc from symbol.
   * @param symbol the symbol
   * @return the reference
   ***************************************************************************/
  public static Reference reference(Symbol symbol)
  {
    return new Reference(symbol.getIdentifier().getName(), symbol);
  }
  
  /****************************************************************************
   * construct reference from value declaration identifier.
   * @param ident the name of the reference
   * @return the reference
   ***************************************************************************/
  public static Reference reference(ValueDeclIdentifier ident)
  {
    return new Reference(ident.getName(), ident.getSymbol());
  }
  
  /****************************************************************************
   * construct reference from type declaration identifier.
   * @param ident the name of the reference
   * @return the reference
   ***************************************************************************/
  public static Reference reference(TypeDeclIdentifier ident)
  {
    return new Reference(ident.getName(), ident.getSymbol());
  }
  
  /****************************************************************************
   * construct value declaration identifier with denoted name.
   * @param name the name of the identifier
   * @return the declaration identifier (with a new and unique internal 
   *   name and an associated value symbol)
   ***************************************************************************/
  public static ValueDeclIdentifier valueDeclIdentifier(String name)
  {
    return new ValueDeclIdentifier(name);
  }
  
  /****************************************************************************
   * construct type declaration identifier with denoted name.
   * @param name the name of the identifier
   * @return the declaration identifier (with a new and unique internal 
   *   name and an associated type symbol)
   ***************************************************************************/
  public static TypeDeclIdentifier typeDeclIdentifier(String name)
  {
    return new TypeDeclIdentifier(name);
  }
  
  /****************************************************************************
   * construct formula declaration identifier with denoted name.
   * @param name the name of the identifier
   * @return the declaration identifier (with a new and unique internal 
   *   name and an associated formula symbol)
   ***************************************************************************/
  public static FormulaDeclIdentifier formulaDeclIdentifier(String name)
  {
    return new FormulaDeclIdentifier(name);
  }
  
  /****************************************************************************
   * construct typed identifier from identifier and type.
   * @param identifier the identifier
   * @param type its type
   * @return the typed identifier
   ***************************************************************************/
  public static TypedIdentifier typedIdentifier(ValueDeclIdentifier identifier, 
    Type type)
  {
    return new TypedIdentifier(identifier, type);
  }
  
  /****************************************************************************
   * construct identifier with value from identifier and value.
   * @param identifier the identifier
   * @param value its value
   * @return the identifier with value
   ***************************************************************************/
  public static ValuedIdentifier valuedIdentifier(Reference identifier, 
    Expression value)
  {
    return new ValuedIdentifier(identifier, value);
  }
  
  /****************************************************************************
   * construct number with denoted value.
   * @param value the digit representation of the number
   * @return the number
   ***************************************************************************/
  public static Number number(String value)
  {
    return new Number(value);
  }
  
  /****************************************************************************
   * construct logical constant with denoted value.
   * @param value the value of the constant
   * @return the constant
   ***************************************************************************/
  public static Logical logical(boolean value)
  {
    return new Logical(value);
  }
  
  /****************************************************************************
   * construct type declaration with denoted name and type.
   * @param name the declared name
   * @param type the type associated to the name
   * @return the type declaration
   ***************************************************************************/
  public static TypeDeclaration typeDeclaration(TypeDeclIdentifier name, 
    Type type)
  {
    return new TypeDeclaration(name, type);
  }
  
  /****************************************************************************
   * construct formula declaration with denoted name and formula.
   * @param name the declared name
   * @param formula the formula associated to the name
   * @return the formula declaration
   ***************************************************************************/
  public static FormulaDeclaration formulaDeclaration
    (FormulaDeclIdentifier name, Expression formula)
  {
    return new FormulaDeclaration(name, formula);
  }
  
  /****************************************************************************
   * construct axiom declaration with denoted name and formula
   * @param name the declared name
   * @param formula the formula associated to the name
   * @return the formula declaration
   ***************************************************************************/
  public static AxiomDeclaration axiomDeclaration
    (FormulaDeclIdentifier name, Expression formula)
  {
    return new AxiomDeclaration(name, formula);
  }
    
  /****************************************************************************
   * construct value declaration with denoted name, type and value.
   * @param name the declared name
   * @param type the type associated to the name
   * @param value the value associated to the name (may be null)
   * @return the type declaration
   ***************************************************************************/
  public static ValueDeclaration valueDeclaration(ValueDeclIdentifier name, 
    Type type, Expression value)
  {
    return new ValueDeclaration(name, type, value);
  }
  
  /****************************************************************************
   * construct function type from types domain and range.
   * @param domain the domain of the type
   * @param range the range of the type
   * @return the function type
   ***************************************************************************/
  public static FunctionType functionType(Type[] domain, Type range)
  {
    return new FunctionType(domain, range);
  }
  
  /****************************************************************************
   * construct atomic type with denoted name.
   * @param name the name of the type
   * @return the atomic type
   ***************************************************************************/
  public static AtomicType atomicType(String name)
  {
    return new AtomicType(name);
  }
  
  /****************************************************************************
   * construct tuple type with denoted base types.
   * @param base the base types
   * @return the tuple type
   ***************************************************************************/
  public static TupleType tupleType(Type[] base)
  {
    return new TupleType(base);
  }
  
  /****************************************************************************
   * construct record type with denoted base types.
   * @param base the base types (field:type pairs)
   * @return the record type
   ***************************************************************************/
  public static RecordType recordType(TypedIdentifier[] base)
  {
    return new RecordType(base);
  }
  
  /****************************************************************************
   * construct array type with denoted index type and base type.
   * @param index the index type of the array
   * @param base the base type of the array
   * @return the array type
   ***************************************************************************/
  public static ArrayType arrayType(Type index, Type base)
  {
    return new ArrayType(index, base);
  }
  
  /****************************************************************************
   * construct subtype with denoted predicate.
   * @param predicate the predicate defining the subtype
   * @return the subtype
   ***************************************************************************/
  public static SubType subType(Expression predicate)
  {
    return new SubType(predicate);
  }
  
  /****************************************************************************
   * construct subrangetype with denoted bounds.
   * @param from the lower bound
   * @param to the upper bound
   * @return the subrangetype
   ***************************************************************************/
  public static SubrangeType subrangeType(Expression from, Expression to)
  {
    return new SubrangeType(from, to);
  }
  
  /****************************************************************************
   * construct let expressions of declarations and base expressions
   * @param declarations the local bindings in the expressions
   * @param base the base expressions
   * @return the let expresssions
   ***************************************************************************/
  public static LetExpression letExpression(ValueDeclaration[] declarations, 
      Expression base)
  {
    return new LetExpression(declarations, base);
  }
  
  /****************************************************************************
   * construct let type of declarations and base type
   * @param declarations the local bindings in the type
   * @param base the base type
   * @return the let type
   ***************************************************************************/
  public static LetType letType(TypeDeclaration[] declarations, Type base)
  {
    return new LetType(declarations, base);
  }
  
  /****************************************************************************
   * construct if-then-else expressions from condition, thenbranch, elsebranch.
   * @param condition the formula deciding which branch is selected
   * @param thenbranch the branch selected if the formula yields true
   * @return thenbranch the branch selected if the formula yields false
   ***************************************************************************/
  public static IfThenElseExpression ifThenElseExpression(
    Expression condition, Expression thenbranch, Expression elsebranch)
  {
    return new IfThenElseExpression(condition, thenbranch, elsebranch);
  }
  
  /****************************************************************************
   * construct negation (not formula) of base formula
   * @param base the base formula
   * @return the negation formula
   ***************************************************************************/
  public static NotFormula notFormula(Expression base)
  {
    return new NotFormula(base);
  }
  
  /****************************************************************************
   * construct negation (minus term) of base term
   * @param base the base term
   * @return the negation term
   ***************************************************************************/
  public static NegationTerm negationTerm(Expression base)
  {
    return new NegationTerm(base);
  }
  
  /****************************************************************************
   * construct tuple term of base terms
   * @param base the array of base terms, base.length > 1
   * @return the tuple term
   ***************************************************************************/
  public static TupleTerm tupleTerm(Expression[] base)
  {
    return new TupleTerm(base);
  }
  
  /****************************************************************************
   * construct record term of base terms
   * @param base the array of base terms, base.length > 1
   * @return the record term
   ***************************************************************************/
  public static RecordTerm recordTerm(ValuedIdentifier[] base)
  {
    return new RecordTerm(base);
  }
  
  /****************************************************************************
   * construct lambda term of variables and base term
   * @param variables the quantified variables
   * @param base the base term
   * @return the lambda term
   ***************************************************************************/
  public static LambdaTerm lambdaTerm(TypedIdentifier[] variables, 
      Expression base)
  {
    return new LambdaTerm(variables, base);
  }
  
  /****************************************************************************
   * construct array term of variables and base term
   * @param variables the quantified variables
   * @param base the base term
   * @return the array term
   ***************************************************************************/
  public static ArrayTerm arrayTerm(TypedIdentifier[] variables, 
      Expression base)
  {
    return new ArrayTerm(variables, base);
  }
  
  /****************************************************************************
   * construct universally quantifed formula of variables and base formula
   * @param variables the quantified variables
   * @param base the base formula
   * @return the universally quantified formula
   ***************************************************************************/
  public static ForallFormula forallFormula(TypedIdentifier[] variables, 
      Expression base)
  {
    return new ForallFormula(variables, base);
  }
  
  /****************************************************************************
   * construct existentially quantifed formula of variables and base formula
   * @param variables the quantified variables
   * @param base the base formula
   * @return the existentially quantified formula
   ***************************************************************************/
  public static ExistsFormula existsFormula(TypedIdentifier[] variables, 
      Expression base)
  {
    return new ExistsFormula(variables, base);
  }
  
  /****************************************************************************
   * construct conjunction with components base1 and base2
   * @param base1 the first component
   * @param base2 the second component
   * @return the formula
   ***************************************************************************/
  public static AndFormula andFormula(Expression base1, Expression base2)
  {
    return new AndFormula(base1, base2);
  }
  
  /****************************************************************************
   * construct disjunction with components base1 and base2
   * @param base1 the first component
   * @param base2 the second component
   * @return the formula
   ***************************************************************************/
  public static OrFormula orFormula(Expression base1, Expression base2)
  {
    return new OrFormula(base1, base2);
  }
  
  /****************************************************************************
   * construct equivalence with components base1 and base2
   * @param base1 the first component
   * @param base2 the second component
   * @return the formula
   ***************************************************************************/
  public static EquivalentFormula equivalentFormula(Expression base1, 
    Expression base2)
  {
    return new EquivalentFormula(base1, base2);
  }
  
  /****************************************************************************
   * construct exclusive disjunction with components base1 and base2
   * @param base1 the first component
   * @param base2 the second component
   * @return the formula
   ***************************************************************************/
  public static NotEquivalentFormula notEquivalentFormula(Expression base1, 
    Expression base2)
  {
    return new NotEquivalentFormula(base1, base2);
  }
  
  /****************************************************************************
   * construct implication with components base1 and base2
   * @param base1 the first component
   * @param base2 the second component
   * @return the formula
   ***************************************************************************/
  public static ImpliesFormula impliesFormula(Expression base1, 
    Expression base2)
  {
    return new ImpliesFormula(base1, base2);
  }
  
  /****************************************************************************
   * construct equality with components base1 and base2
   * @param base1 the first component
   * @param base2 the second component
   * @return the expression
   ***************************************************************************/
  public static EqualsFormula equalsFormula(Expression base1, 
    Expression base2)
  {
    return new EqualsFormula(base1, base2);
  }
  
  /****************************************************************************
   * construct inequality with components base1 and base2
   * @param base1 the first component
   * @param base2 the second component
   * @return the formula
   ***************************************************************************/
  public static NotEqualsFormula notEqualsFormula(Expression base1, 
    Expression base2)
  {
    return new NotEqualsFormula(base1, base2);
  }
  
  /****************************************************************************
   * construct binary sum with components base1 and base2
   * @param base1 the first component
   * @param base2 the second component
   * @return the term
   ***************************************************************************/
  public static PlusTerm plusTerm(Expression base1, Expression base2)
  {
    return new PlusTerm(base1, base2);
  }
  
  /****************************************************************************
   * construct difference with components base1 and base2
   * @param base1 the first component
   * @param base2 the second component
   * @return the term
   ***************************************************************************/
  public static MinusTerm minusTerm(Expression base1, Expression base2)
  {
    return new MinusTerm(base1, base2);
  }
  
  /****************************************************************************
   * construct binary product with components base1 and base2
   * @param base1 the first component
   * @param base2 the second component
   * @return the term
   ***************************************************************************/
  public static TimesTerm timesTerm(Expression base1, Expression base2)
  {
    return new TimesTerm(base1, base2);
  }
  
  /****************************************************************************
   * construct quotient with components base1 and base2
   * @param base1 the first component
   * @param base2 the second component
   * @return the term
   ***************************************************************************/
  public static DividesTerm dividesTerm(Expression base1, Expression base2)
  {
    return new DividesTerm(base1, base2);
  }
  
  /****************************************************************************
   * construct power term with components base1 and base2
   * @param base1 the first component
   * @param base2 the second component
   * @return the term
   ***************************************************************************/
  public static PowerTerm powerTerm(Expression base1, Expression base2)
  {
    return new PowerTerm(base1, base2);
  }
  
  /****************************************************************************
   * construct less-than formula with components base1 and base2
   * @param base1 the first component
   * @param base2 the second component
   * @return the formula
   ***************************************************************************/
  public static LessFormula lessFormula(Expression base1, Expression base2)
  {
    return new LessFormula(base1, base2);
  }
  
  /****************************************************************************
   * construct less-than-or-equal formula with components base1 and base2
   * @param base1 the first component
   * @param base2 the second component
   * @return the formula
   ***************************************************************************/
  public static LessEqualFormula lessEqualFormula(Expression base1, 
    Expression base2)
  {
    return new LessEqualFormula(base1, base2);
  }
  
  /****************************************************************************
   * construct greater-than formula with components base1 and base2
   * @param base1 the first component
   * @param base2 the second component
   * @return the formula
   ***************************************************************************/
  public static GreaterFormula greaterFormula(Expression base1, 
    Expression base2)
  {
    return new GreaterFormula(base1, base2);
  }
  
  /****************************************************************************
   * construct greater-than-or-equal formula with components base1 and base2
   * @param base1 the first component
   * @param base2 the second component
   * @return the formula
   ***************************************************************************/
  public static GreaterEqualFormula greaterEqualFormula(Expression base1, 
    Expression base2)
  {
    return new GreaterEqualFormula(base1, base2);
  }

  /****************************************************************************
   * make selector from index expression base
   * @param base the selector base
   * @return the selector
   ***************************************************************************/
  public static SelectorIndex selectorIndex(Expression base)
  {
    return new SelectorIndex(base);
  }
  
  /****************************************************************************
   * make selector from number base
   * @param base the selector base
   * @return the selector
   ***************************************************************************/
  public static SelectorNumber selectorNumber(Number base)
  {
    return new SelectorNumber(base);
  }
  
  /****************************************************************************
   * make selector from identifier base
   * @param base the selector base
   * @return the selector
   ***************************************************************************/
  public static SelectorIdentifier selectorIdentifier(Reference base)
  {
    return new SelectorIdentifier(base);
  }
  
  /****************************************************************************
   * construct component selection from base1 and base2
   * @param base1 the compound object
   * @param base2 the selector
   * @return the term
   ***************************************************************************/
  public static SelectionTerm selectionTerm(Expression base1, Selector base2)
  {
    return new SelectionTerm(base1, base2);
  }
  
  /****************************************************************************
   * construct component update by index selection from base, base2, and base3
   * @param base1 the compound object
   * @param base2 the sequence of selectors
   * @param value the update value
   * @return the term
   ***************************************************************************/
  public static UpdateTerm updateTerm(Expression base1, 
    Selector[] base2, Expression value)
  {
    return new UpdateTerm(base1, base2, value);
  }
  
  /****************************************************************************
   * construct application of base1 to base2
   * @param base1 the function/predicate
   * @param base2 the sequence of arguments
   * @return the expression
   ***************************************************************************/
  public static ApplicationExpression applicationExpression(
    Expression base1, Expression[] base2)
  {
    return new ApplicationExpression(base1, base2);
  }
}