// ----------------------------------------------------------------------------
// UniqueNameTable.java
// Table used to create unique names from given names.
// $Id: UniqueNameTable.java,v 1.2 2005/10/24 19:15:06 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ---------------------------------------------------------------------------
package fmrisc.Syntax;

import fmrisc.Semantics.*;
import java.util.*;

/**
 * 
 *  Table used to create unique names from given names.
 *
 */
public final class UniqueNameTable
{
  // keys are String objects, values are Int objects
  private HashMap table = new HashMap(); 
  
  // container for integer values
  private final static class Int
  {
    int value;
    Int(int value) { this.value = value; }
  }
  
  /****************************************************************************
   * Create empty name table.
   ***************************************************************************/
  public UniqueNameTable()
  {
  }
  
  /****************************************************************************
   * Create name table such that no generated name conflicts with any 
   * value identifier in the given environment.
   * @param env the environment whose value identifiers are already used.
   ***************************************************************************/
  public UniqueNameTable(Environment env)
  {
    Iterator iterator = env.getValueSymbols().iterator();
    while (iterator.hasNext())
    {
      ValueSymbol symbol = (ValueSymbol)iterator.next();
      String name = symbol.getIdentifier().getName();
      int ext = getExtension(name);
      if (ext == -1) continue;
      String base = getBaseName(name);
      Int number = (Int)(table.get(base));
      if (number == null)
        table.put(base, new Int(ext+1));
      else
        number.value = Math.max(number.value, ext+1);
    }
  }
  
  /****************************************************************************
   * Create unique identifier name from given name.
   * @param name of an identifier
   * @return a unique version of name not in use anywhere else as a unique id.
   **************************************************************************/
  public String getUniqueName(String name)
  {
    name = getBaseName(name);
    Int number = (Int)(table.get(name));
    if (number != null)
    {
      int v = number.value;
      number.value++;
      return name + "_" + Integer.toString(v);
    }
    else
    {
      number = new Int(1);
      table.put(name, number);
      return name + "_0";
    }
  }
  
  /***************************************************************************
   * Get base name from name.
   * @param name a name (possible unique as returned by getUniqueName()).
   * @return the original base name (which may be the name itself).
   **************************************************************************/
  public static String getBaseName(String name)
  {
    int pos = name.lastIndexOf('_');
    if (pos == -1) return name;
    return name.substring(0, pos);
  }
  
  /***************************************************************************
   * Get extension from name
   * @param name a name (possible unique as returned by getUniqueName()).
   * @return the extension (non-negative, -1 indicates name has no extension)
   **************************************************************************/
  private static int getExtension(String name)
  {
    int pos = name.lastIndexOf('_');
    if (pos == -1) return -1;
    String s = name.substring(pos+1);
    try
    {
      return Integer.parseInt(s);
    }
    catch (NumberFormatException e)
    {
      return -1;
    }
  }
}
